searchNodes=[{"doc":"The step specification. There are two kind of steps: function steps and module steps. Function steps A function step is any function that receives a token and a set of options and returns a token. Its type signature must be: ( Pluggable.Token . t , Pluggable . opts ) :: Pluggable.Token . t Module steps A module step is an extension of the function step. It is a module that must export: a call/2 function with the signature defined above an init/1 function which takes a set of options and initializes it. The result returned by init/1 is passed as second argument to call/2 . Note that init/1 may be called during compilation and as such it must not return pids, ports or values that are specific to the runtime. The API expected by a module step is defined as a behaviour by the Pluggable module (this module). Examples Here's an example of a function step: def json_header_step ( token , _opts ) do My.Token . put_data ( token , &quot;some_data&quot; ) end Here's an example of a module step: defmodule PutSomeData do def init ( opts ) do opts end def call ( token , _opts ) do My.Token . put_data ( token , &quot;some_data&quot; ) end end The Pluggable Step pipeline The Pluggable.StepBuilder module provides conveniences for building pluggable step pipelines.","ref":"Pluggable.html","title":"Pluggable","type":"behaviour"},{"doc":"","ref":"Pluggable.html#c:call/2","title":"Pluggable.call/2","type":"callback"},{"doc":"","ref":"Pluggable.html#c:init/1","title":"Pluggable.init/1","type":"callback"},{"doc":"Run a series of pluggable steps at runtime. The steps given here can be either a tuple, representing a module step and their options, or a simple function that receives a token and returns a token. If any of the steps halt, the remaining steps are not invoked. If the given token was already halted, none of the steps are invoked either. While Pluggable.StepBuilder works at compile-time, this is a straight-forward alternative that works at runtime. Examples Pluggable . run ( token , [ { My.Step , [ ] } , &amp; IO . inspect / 1 ] ) Options :log_on_halt - a log level to be used if a pipeline halts","ref":"Pluggable.html#run/3","title":"Pluggable.run/3","type":"function"},{"doc":"","ref":"Pluggable.html#t:opts/0","title":"Pluggable.opts/0","type":"type"},{"doc":"Conveniences for building pipelines. defmodule MyApp do use Pluggable.StepBuilder step SomeLibrary.Logger step :hello , upper : true # A function from another module can be plugged too, provided it&#39;s # imported into the current module first. import AnotherModule , only : [ interesting_step : 2 ] step :interesting_step def hello ( token , opts ) do body = if opts [ :upper ] , do : &quot;WORLD&quot; , else : &quot;world&quot; send_resp ( token , 200 , body ) end end Multiple steps can be defined with the step/2 macro, forming a pipeline. The steps in the pipeline will be executed in the order they've been added through the step/2 macro. In the example above, SomeLibrary.Logger will be called first and then the :hello function step will be called on the resulting token. Options When used, the following options are accepted by Pluggable.StepBuilder : :init_mode - the environment to initialize the step's options, one of :compile or :runtime . Defaults :compile . :log_on_halt - accepts the level to log whenever the request is halted :copy_opts_to_assign - an atom representing an assign. When supplied, it will copy the options given to the Step initialization to the given token assign step behaviour Internally, Pluggable.StepBuilder implements the Pluggable behaviour, which means both the init/1 and call/2 functions are defined. By implementing the Pluggable API, Pluggable.StepBuilder guarantees this module is a pluggable step and can be run via Pluggable.run/3 or used as part of another pipeline. Overriding the default Pluggable API functions Both the init/1 and call/2 functions defined by Pluggable.StepBuilder can be manually overridden. For example, the init/1 function provided by Pluggable.StepBuilder returns the options that it receives as an argument, but its behaviour can be customized: defmodule StepWithCustomOptions do use Pluggable.StepBuilder step SomeLibrary.Logger def init ( opts ) do opts end end The call/2 function that Pluggable.StepBuilder provides is used internally to execute all the steps listed using the step macro, so overriding the call/2 function generally implies using super in order to still call the step chain: defmodule StepWithCustomCall do use Pluggable.StepBuilder step SomeLibrary.Logger step SomeLibrary.AddMeta def call ( token , opts ) do token |&gt; super ( opts ) # calls SomeLibrary.Logger and SomeLibrary.AddMeta |&gt; assign ( :called_all_steps , true ) end end Halting a pluggable step pipeline A pluggable step pipeline can be halted with Pluggable.Token.halt/1 . The builder will prevent further steps downstream from being invoked and return the current token. In the following example, the SomeLibrary.Logger step never gets called: defmodule StepUsingHalt do use Pluggable.StepBuilder step :stopper step SomeLibrary.Logger def stopper ( token , _opts ) do halt ( token ) end end","ref":"Pluggable.StepBuilder.html","title":"Pluggable.StepBuilder","type":"module"},{"doc":"Compiles a pluggable step pipeline. Each element of the pluggable step pipeline (according to the type signature of this function) has the form: { step_name , options , guards } Note that this function expects a reversed pipeline (with the last step that has to be called coming first in the pipeline). The function returns a tuple with the first element being a quoted reference to the token and the second element being the compiled quoted pipeline. Examples Pluggable.StepBuilder . compile ( env , [ { SomeLibrary.Logger , [ ] , true } , # no guards, as added by the Pluggable.StepBuilder.step/2 macro { SomeLibrary.AddMeta , [ ] , quote ( do : a when is_binary ( a ) ) } ] , [ ] )","ref":"Pluggable.StepBuilder.html#compile/3","title":"Pluggable.StepBuilder.compile/3","type":"function"},{"doc":"A macro that stores a new step. opts will be passed unchanged to the new step. This macro doesn't add any guards when adding the new step to the pipeline; for more information about adding steps with guards see compile/3 . Examples step SomeLibrary.Logger # step module step :foo , some_options : true # step function","ref":"Pluggable.StepBuilder.html#step/2","title":"Pluggable.StepBuilder.step/2","type":"macro"},{"doc":"","ref":"Pluggable.StepBuilder.html#t:step/0","title":"Pluggable.StepBuilder.step/0","type":"type"},{"doc":"Token protocol to be used with the Pluggable and Pluggable.Builder modules. When implementing pipelines using this library, a token holds the state and is passed on from step to step within the pipeline. Deriving Pluggable.Token The simplest way to use this library is to define a token module which derives Pluggable.Token and defines a struct which, among others defines the keys: :halted - the boolean status on whether the pipeline was halted :assigns - shared user data as a map Example: defmodule MyPipeline.Token do @derive Pluggable.Token defstruct [ halted : false , assigns : %{ } , # other state ] end If the fields holding these two states are named differently, pass the fields as options to @derive : defmodule MyPipeline.Token do @derive { Pluggable.Token , halted_key : :stopped , assigns_key : :shared_state } defstruct [ stopped : false , shared_state : %{ } , # other state ] end Implementing Pluggable.Token Pluggable.Token can be implemented. The following is the default implementation when deriving Pluggable.Token defmodule MyPipeline.Token do defstruct [ halted : nil , assigns : %{ } , # other state ] end defimpl Pluggable.Token , for : MyPipeline.Token do def halted? ( token ) , do : token . halted def halt ( token ) , do : %{ token | halted : true } def assign ( % MyPipeline.Token { assigns : assigns } = token , key , value ) when is_atom ( key ) do %{ token | assigns : Map . put ( assigns , key , value ) } end end","ref":"Pluggable.Token.html","title":"Pluggable.Token","type":"protocol"},{"doc":"Assigns a value to a key in the shared user data map.","ref":"Pluggable.Token.html#assign/3","title":"Pluggable.Token.assign/3","type":"function"},{"doc":"Halts the Pluggable pipeline by preventing further steps downstream from being invoked. See the docs for Pluggable.Builder for more information on halting a Pluggable pipeline.","ref":"Pluggable.Token.html#halt/1","title":"Pluggable.Token.halt/1","type":"function"},{"doc":"Returns the boolean status on whether the pipeline was halted","ref":"Pluggable.Token.html#halted?/1","title":"Pluggable.Token.halted?/1","type":"function"},{"doc":"","ref":"Pluggable.Token.html#t:t/0","title":"Pluggable.Token.t/0","type":"type"},{"doc":"Pluggable helps to define Plug like pipelines but with arbitrary tokens. The library comes with almost exact copies of the module Plug and Plug.Builder . However, instead of passing around a %Plug.Conn{} struct, this library passes around a Token you define in your project.","ref":"readme.html","title":"Pluggable","type":"extras"},{"doc":"Most of the code in this module was copied from the :plug library so credits go to the creators and maintainers of :plug .","ref":"readme.html#credits","title":"Pluggable - Credits","type":"extras"},{"doc":"The package can be installed by adding pluggable to your list of dependencies in mix.exs : def deps do [ { :pluggable , &quot;~&gt; 1.0&quot; } ] end","ref":"readme.html#installation","title":"Pluggable - Installation","type":"extras"},{"doc":"To use this library, you first have to define your token. Once that is done, use Pluggable.StepBuilder to build steps and pipelines. Deriving Pluggable.Token The easiest way to define a token is to create a module which derives Pluggable.Token and defines a struct which, among others defines the keys: :halted - the boolean status on whether the pipeline was halted :assigns - shared user data as a map Example: defmodule MyPipeline.Token do @derive Pluggable.Token defstruct [ halted : false , assigns : %{ } , # other state ] end If the fields holding these two states are named differently, pass the fields as options to @derive : defmodule MyPipeline.Token do @derive { Pluggable.Token , halted_key : :stopped , assigns_key : :shared_state } defstruct [ stopped : false , shared_state : %{ } , # other state ] end Implementing Pluggable.Token Pluggable.Token can be implemented. The following is the default implementation when deriving Pluggable.Token defmodule MyPipeline.Token do defstruct [ halted : nil , assigns : %{ } , # other state ] end defimpl Pluggable.Token , for : MyPipeline.Token do def halted? ( token ) , do : token . halted def halt ( token ) , do : %{ token | halted : true } def assign ( % MyPipeline.Token { assigns : assigns } = token , key , value ) when is_atom ( key ) do %{ token | assigns : Map . put ( assigns , key , value ) } end end","ref":"readme.html#usage","title":"Pluggable - Usage","type":"extras"},{"doc":"Pluggable.StepBuilder works just like Plug.Builder . See the module documentation for instructions.","ref":"readme.html#building-pipelines","title":"Pluggable - Building Pipelines","type":"extras"},{"doc":"All notable changes to this project will be documented in this file. The format is based on Keep a Changelog , and this project adheres to Semantic Versioning .","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"","ref":"changelog.html#unreleased","title":"Changelog - Unreleased","type":"extras"},{"doc":"This is the first release.","ref":"changelog.html#1-0-0-2022-10-12","title":"Changelog - [1.0.0] - 2022-10-12","type":"extras"}]